#!/usr/bin/python

import os
from contextlib import contextmanager

from yawrap._engine import Doc
from yawrap._sourcer import HEAD, BODY_BEGIN, BODY_END, _Resource
from yawrap.utils import assert_keys_not_in, make_place

DEFAULT_SVG_TAG_ATTRIBUTES = dict(xmlns="http://www.w3.org/2000/svg", version="1.1")


@contextmanager
def svg_structure(painter,
                  svg_tag_attributes=None,
                  svg_styles_as_str="",
                  *args,
                  **kwargs):

    kwargs.update(svg_tag_attributes or DEFAULT_SVG_TAG_ATTRIBUTES)
    with painter.tag('svg', *args, **kwargs):
        if svg_styles_as_str:
            with painter.tag('style', type="text/css"):
                painter.cdata(svg_styles_as_str)
        yield


class Yawrap(Doc):
    resources = []
    html_d = dict(lang="en-US")
    meta_d = [dict(charset="UTF-8")]
    svg_d = DEFAULT_SVG_TAG_ATTRIBUTES

    def __init__(self, target_file, title='', parent=None):

        super(Yawrap, self).__init__()
        self._target_file = target_file
        self.title = title
        self._parent = parent
        self._target_dir = os.path.dirname(target_file)
        self._additional_resources = []

    @contextmanager
    def local_link(self, target, *args, **kwargs):
        rel_location = os.path.relpath(target, self._target_dir)
        assert_keys_not_in('href', args, kwargs)
        with self.tag('a', href=rel_location, *args, **kwargs):
            yield rel_location

    def render(self):
        """ Saves page in current state to target file. """
        raw_html = self._render_page()
        with open(make_place(self._target_file), 'wt') as ff:
            ff.write(raw_html)

    def getvalue(self):
        return self._render_page()

    @contextmanager
    def svg(self, svg_styles_as_str='', *args, **svg_tag_attributes):
        svg_tag_attributes.update(self.svg_d)
        with svg_structure(self, svg_tag_attributes, svg_styles_as_str, *args, **svg_tag_attributes):
            yield

    def add(self, js_or_css_resource):
        assert isinstance(js_or_css_resource, _Resource), "Bad ussage, expected CSS or JS" \
                                                          " resource definition, got %s." % type(
            js_or_css_resource).__name__
        self._additional_resources.append(js_or_css_resource)

    def _get_rel_path(self, target_local_file):
        return os.path.relpath(os.path.abspath(target_local_file), self._target_dir)

    @contextmanager
    def _html_page_structure(self, page_doc):

        page_doc.asis('<!doctype html>')
        with page_doc.tag('html', **self.html_d):
            with page_doc.tag('head'):
                for meta in self.meta_d:
                    page_doc.stag('meta', **meta)
                if self.title:
                    with page_doc.tag('title'):
                        page_doc.text(self.title)

                for resource in self.resources + self._additional_resources:
                    resource.visit(page_doc, self, HEAD)

            with page_doc.tag('body'):
                for resource in self.resources + self._additional_resources:
                    resource.visit(page_doc, self, BODY_BEGIN)
                """ at this moment yawrap will place its html here """
                yield
                for resource in self.resources + self._additional_resources:
                    resource.visit(page_doc, self, BODY_END)

    def _render_page(self):
        page_doc = Doc()
        with self._html_page_structure(page_doc):
            page_doc._clone_children(self)
        return page_doc.getvalue()

    def _get_root(self):
        return self

    def get_root_dir(self):
        return self._get_root()._target_dir



from contextlib import contextmanager

from yawrap.six import str_types

"""
    The purpose of this module is to replace dependency to yattag with own machinery.
    The Doc class shares same interface with yattag. Cannot guarantee that, but in vast most
    of cases this module's code fully substitutes yattag.SimpleDoc class.

    The main reason - to make own code that supplements well known package is that:
    - the well known package doesn't create indentation on the fly.
    - the well known package doesn't like to change. New features are not welcome.
    - because no dependency and pure pythonic code are nice
"""


class IndentableLine(object):
    __slots__ = ("text", "indent_level")
    indenter = "  "

    def __init__(self, text):
        self.text = text
        self.indent_level = 0

    def increase_indent(self):
        self.indent_level += 1

    def __str__(self):
        indent = self.indenter * self.indent_level
        return "%s%s" % (indent, self.text)


class NonindentableLine(IndentableLine):
    indenter = ""


ATTRIBUTE_SUBSTITUTES = {
    'klass': 'class',
    'Class': 'class',
    'class_': 'class',
    'fill_opacity': 'fill-opacity',
    'stroke_width': 'stroke-width',
    'stroke_dasharray': ' stroke-dasharray',
    "stroke_opacity": "stroke-opacity",
    "stroke_dashoffset": "stroke-dashoffset",
    "stroke_linejoin": "stroke-linejoin",
    "stroke_miterlimit": "stroke-miterlimit",
}

DEFAULT_ATTRIBUTE_ORDER = {
    "class": 0,
    "id": 1,
    "name": 2,
    "src": 3,
    "for": 4,
    "type": 5,
    "href": 6,
    "value": 7,
    "default rank for other attributes": 10,
    "title": 101,
    "alt": 102,
    "style": 103,
}


class Tag(object):
    __slots__ = ("tag_name", "attributes", "children")
    attribute_precedence_rank = DEFAULT_ATTRIBUTE_ORDER
    max_line_width = 100

    def __init__(self, tag_name, *args, **kwargs):
        self.tag_name = tag_name
        self.attributes = dict(self._process_attributes(args, kwargs))
        self.children = []

    def __str__(self):
        return "\n".join(str(l) for l in self._build_lines())

    @classmethod
    def _process_attributes(cls, args, kwargs):
        for arg in args:
            if isinstance(arg, tuple):
                if len(arg) != 2:
                    raise ValueError("Attribute argument must be tuple of 2 elements (name, value).")
                yield arg
            elif isinstance(arg, str_types):
                yield (arg, None)
            else:
                raise ValueError("Couldn't make an attribute & value pair out of %r." % arg)

        for key, value in kwargs.items():
            unslugged_key = ATTRIBUTE_SUBSTITUTES.get(key, key)
            yield unslugged_key, value

    @classmethod
    def _form_attributes(cls, attributes_dict):
        assert isinstance(attributes_dict, dict)

        def sorting_function(element):
            attribute_name = element[0]
            default_rank = cls.attribute_precedence_rank["default rank for other attributes"]
            attribute_rank = cls.attribute_precedence_rank.get(attribute_name, default_rank)
            return attribute_rank, attribute_name

        def collect_attributes():
            for key, value in sorted(attributes_dict.items(), key=sorting_function):
                if value is None:
                    yield " %s" % key
                else:
                    yield " %s=%r" % (key, value)

        return ''.join(list(collect_attributes()))

    def _update_attributes(self, *args, **kwargs):
        new_attrs = dict(self._process_attributes(args, kwargs))
        self.attributes.update(new_attrs)

    def _build_lines(self):
        begin_tag = "<%s%s>" % (self.tag_name, self._form_attributes(self.attributes))
        end_tag = "</%s>" % self.tag_name
        children_lines = [line for child in self.children for line in child._build_lines()]

        if len(children_lines) <= 1:
            # try to fit it in single line
            parts = [begin_tag] + [str(l) for l in children_lines] + [end_tag]
            total_length = sum(map(len, parts))
            if total_length <= self.max_line_width:
                return [IndentableLine("".join(parts))]

        for child_line in children_lines:
            child_line.increase_indent()
        return [IndentableLine(begin_tag)] + children_lines + [IndentableLine(end_tag)]


class SingleTag(Tag):
    """ E.g.: <br />, <hr /> or <img src="#" alt="" /> """

    def _build_lines(self):
        tag_body = "<%s%s />" % (self.tag_name, self._form_attributes(self.attributes))
        return [IndentableLine(tag_body)]


class Text(object):
    __slots__ = ("_text",)

    def __init__(self, text):
        self._text = text

    def _build_lines(self):
        return [IndentableLine(l) for l in str(self).split("\n")]

    def __str__(self):
        return self._text


class PlainText(Text):

    def _build_lines(self):
        return [NonindentableLine(str(self))]


class Comment(Text):
    def __str__(self):
        return "<!-- %s -->" % self._text


class CdataTag(PlainText):

    def __str__(self):
        return "<![CDATA[%s]]>" % self._text


class Doc(object):
    class DocumentRootTag(object):
        __slots__ = ("children",)

        def __init__(self):
            self.children = []

    def __init__(self):
        self._root_tag = Doc.DocumentRootTag()
        self.__tag_stack = [self._root_tag]

    def __str__(self):
        return "".join(str(c) for c in self._root_tag.children)

    @property
    def _top_element(self):
        return self.__tag_stack[-1]

    @property
    def __root_element(self):
        return self.__tag_stack[0]

    @property
    def _attributes(self):
        return self._top_element.attributes

    @property
    def _children(self):
        return self._top_element.children

    def _clone_children(self, other_doc):
        assert isinstance(other_doc, Doc)
        self._children.extend(other_doc.__root_element.children)

    @contextmanager
    def tag(self, tag_name, *args, **kwargs):
        tag = Tag(tag_name, *args, **kwargs)
        self._children.append(tag)
        self.__tag_stack.append(tag)
        try:
            yield
        finally:
            self.__tag_stack.pop()

    def text(self, *text):
        self._children.extend([Text(t) for t in text])

    def line(self, tag_name, content="", *args, **kwargs):
        tag = Tag(tag_name, *args, **kwargs)
        tag.children.append(Text(content))
        self._children.append(tag)

    def stag(self, tag_name, *args, **kwargs):
        tag = SingleTag(tag_name, *args, **kwargs)
        self._children.append(tag)

    def tagtext(self):  # pragma: no cover
        return self, self.tag, self.text

    def ttl(self):  # pragma: no cover
        return self, self.tag, self.text, self.line

    def getvalue(self):
        return str(self)

    def comment(self, text):
        self._children.append(Comment(text))

    def cdata(self, raw_text):
        self._children.append(CdataTag(raw_text))

    def asis(self, raw_text):
        """ add the text, whatever is there """
        self._children.append(PlainText(raw_text))

    def nl(self):
        """ adds blank line to resulting html document """
        self.asis("")

    def attr(self, *args, **kwargs):
        """ update attribures of current element """
        self._top_element._update_attributes(*args, **kwargs)

    def classes(self):
        top_element = self._top_element
        assert not isinstance(top_element, self.DocumentRootTag), "Root element has no classes."
        return top_element.attributes.get("class", "").split()

    def add_class(self, names):
        """ adds class names (space-separated words) to current element """
        assert isinstance(names, str_types)
        classes = self.classes()

        for name in names.split():
            if name not in classes:
                classes.append(name)

        self._attributes.update({"class": " ".join(classes)})

    def discard_class(self, names):
        """ removes class names (space-separated words) to current element """
        assert isinstance(names, str_types)
        classes = self.classes()
        for name in names.split():
            if name in classes:
                classes.pop(classes.index(name))
        if classes:
            self._attributes.update({"class": " ".join(classes)})
        else:
            self._attributes.pop("class", None)

    def toggle_class(self, names):
        """ toggles class names (space-separated words) to current element """
        already_processed = []
        for name in names.split():
            if name not in already_processed:
                if name in self.classes():
                    self.discard_class(name)
                else:
                    self.add_class(name)
                already_processed.append(name)



import os
import weakref
from collections import namedtuple
from contextlib import contextmanager

from ._engine import Doc
from ._yawrap import Yawrap
from .six import str_types
from .utils import assert_keys_not_in

NavEntry = namedtuple('NavEntry', 'element, bookmarks, children')

BASIC_NAV_CSS = """\
.nav_main_panel {
    padding: 6px;
}
.nav_group_div {
    padding: 0px 8px;
    margin: 6px;
}
.main_content_body {
    padding: 3px;
}
"""


class NavedYawrap(Yawrap):
    css = BASIC_NAV_CSS

    def __init__(self, target_file, title='', parent=None, nav_title=''):
        super(NavedYawrap, self).__init__(target_file, title, parent)
        self.nav_title = nav_title or title
        self._subs = []
        self._bookmarks = []

    def sub(self, target_file, title=''):
        sub_ = type(self)(target_file, title, weakref.ref(self))
        self._subs.append(sub_)
        return sub_

    @contextmanager
    def bookmark(self, id_, name_in_nav='', type_='div', *args, **kwargs):
        """ as regular doc.tag, but also manages navigation stuff """
        assert_keys_not_in('id', args, kwargs)
        assert id_ and isinstance(id_, str_types), "Invalid id: '%s'" % id_
        assert id_ not in map(lambda x: x[0], self._bookmarks), "Bookmark ids collision. '%s' is already defined" % id_
        name_in_nav = name_in_nav or id_
        id_ = id_.replace(' ', '_')
        self._bookmarks.append((id_, name_in_nav))
        with self.tag(type_, *args, id=id_, **kwargs):
            yield

    def render_all_files(self):
        self.render()
        for sub in self._subs:
            sub.render_all_files()

    def _render_page(self):
        page_doc = Doc()
        with self._html_page_structure(page_doc):
            if self._subs or self._parent:
                self._insert_nav(page_doc)
            with page_doc.tag('main', klass='main_content_body'):
                page_doc._clone_children(self)
        return page_doc.getvalue()

    def _get_root(self):
        if self._parent:
            return self._parent()._get_root()
        return self

    def _insert_nav(self, doc):
        nav_structure = self._get_nav_structure()
        with doc.tag('nav', klass='nav_main_panel'):
            self._render_nav_subs(nav_structure, doc)

    def _get_nav_structure(self, sub=None):
        current = sub or self._get_root()
        its_children = tuple(child._get_nav_structure(child) for child in current._subs)
        return NavEntry(current, current._bookmarks, its_children)

    def _render_nav_subs(self, structure_element, doc):
        current = structure_element.element
        link = os.path.relpath(current._target_file, self._target_dir)

        with doc.tag('div', klass='nav_group_div'):
            if current == self:
                doc.attr(klass='nav_group_div active')

            with doc.tag('div', klass='nav_page'):
                with doc.tag('a', href=link, klass='nav_page_link'):
                    if current._target_file == self._target_file:
                        doc.attr(klass='active')
                    doc.text(current.nav_title or link)

                if current == self:
                    doc.attr(klass='nav_page with_bookmarks')
                    for bookmark, bookmark_name in structure_element.bookmarks:
                        bookmark_link = "%s#%s" % (link, bookmark)
                        with doc.tag('a', klass='nav_bookmark_link', href=bookmark_link):
                            doc.text(bookmark_name)

            for sub in structure_element.children:
                self._render_nav_subs(sub, doc)



# flake8: noqa
import sys

if sys.version < '3':
    str_types = (str, unicode)
    from itertools import ifilter
    from urllib2 import urlopen
    from urlparse import urlparse

else:
    ifilter = filter
    str_types = str
    from urllib.request import urlopen
    from urllib.parse import urlparse



"""
    There are some possibilities when you want to use CSS or JS scripts in a html page.

    A. When the content is on the web, specified by an url - you may want to:
        1. download it (while the page is being built), save it and reference it from a directory
        relative to target page (several pages may wont like to share the file)
        2. embed it in the target page (it's better for a single-file document type)
        3. just reference it by the url (will require web access for each view)

    B. When the content is in a local file - you may want to:
        1. export it to directory relative to a target document or
        2. embed it in the document (CSS in head, JS anywhere)

    B'. When the content is in loacal python string - you may want to:
        1. export it to directory relative to a target document or
        2. embed it in the document (CSS in head, JS anywhere)
"""

import os
import posixpath
from contextlib import closing

from .six import urlopen, urlparse, str_types
from .utils import form_css, make_place, is_url

HEAD = "head"
BODY_END = "body_end"
BODY_BEGIN = "body_begin"
PLACEMENT_OPTIONS = [HEAD, BODY_BEGIN, BODY_END]
RAISE_ON_DOWNLOAD_FAIL = True


def warn_(text):
    print(text)


class _Resource(object):
    """
        Class that provides methods of resource content's acquisition.
        The from_url and from_file classmethods are supposed to serve alternative contructors.
    """

    def __init__(self, read_method, file_name):
        self.read_method = read_method
        self.file_name = file_name

    @classmethod
    def from_url(cls, url, placement=HEAD):
        """ Provide content of the resource from the web. """
        assert url, "Bad argument: %s" % url
        assert is_url(url), "That doesn't seem to be a valid url: %s" % url
        assert issubclass(cls, _DocumentVisitor), "You messed up."

        def read_method():
            return cls._download(url)

        file_name = posixpath.basename(urlparse(url).path)
        return cls(read_method, placement, file_name)

    @classmethod
    def from_file(cls, file_path, placement=HEAD):
        """ Provide content of the resource from a local file. """
        assert file_path, "Bad argument: %s" % file_path
        file_name = os.path.basename(file_path)

        def read_method():
            return cls._read_file(file_path)

        assert issubclass(cls, _DocumentVisitor), "You messed up."
        return cls(read_method, placement, file_name)

    @staticmethod
    def _download(url):
        assert is_url(url), "That doesn't seem to be a valid url: %s" % url
        try:
            with closing(urlopen(url, timeout=3)) as response:
                return response.read()
        except Exception as e:  # pragma: no cover
            if RAISE_ON_DOWNLOAD_FAIL:
                raise
            else:
                print("unable to download: %s\n%s" % (url, e))
                return ""

    @staticmethod
    def _read_file(file_path):
        assert os.path.isfile(file_path), "File doesn't exist: %s" % file_path
        with open(file_path, "rt") as ff:
            return ff.read()


class _DocumentVisitor(_Resource):

    def __init__(self, read_function_or_str, placement=HEAD, file_name=None):
        self._check_placement(placement)
        self._placement = placement
        read_method = self._form_read_method(read_function_or_str)
        _Resource.__init__(self, read_method, file_name)

    @classmethod
    def _form_read_method(cls, read_function_or_str):
        if isinstance(read_function_or_str, str_types):
            def read_method():
                return read_function_or_str

            return read_method
        else:
            return read_function_or_str

    def _check_placement(self, placement):
        assert (placement in PLACEMENT_OPTIONS), "Wrong placement value: %s" % placement

    def _placement_match(self, placement):
        return self._placement == placement

    def visit(self, page_doc, yawrap_instance, placement):
        if self._placement_match(placement):
            self.act(page_doc, yawrap_instance)

    @classmethod
    def act(self):
        raise ValueError("You messed up. That shouldn't happen.")


class _JsResource(_DocumentVisitor):
    type_ = "text/javascript"

    @classmethod
    def link(cls, page_doc, href):
        with page_doc.tag('script', src=href):
            pass

    @classmethod
    def embed(cls, page_doc, content):
        with page_doc.tag('script', type=cls.type_):
            page_doc.asis(content)


class _CssResource(_DocumentVisitor):
    rel = "stylesheet"
    type_ = "text/css"

    @classmethod
    def link(cls, page_doc, href):
        page_doc.stag('link', rel=cls.rel, type=cls.type_, href=href)

    @classmethod
    def embed(cls, page_doc, content):
        with page_doc.tag('style'):
            page_doc.asis(content)

    @classmethod
    def _form_read_method(cls, function_or_str_or_dict):
        if isinstance(function_or_str_or_dict, str_types):
            def read_method():
                return function_or_str_or_dict

            return read_method
        elif isinstance(function_or_str_or_dict, dict):
            def read_method():
                return form_css(function_or_str_or_dict)

            return read_method
        else:
            return function_or_str_or_dict

    def _check_placement(self, placement):
        _DocumentVisitor._check_placement(self, placement)
        if placement != HEAD:
            raise TypeError("Cannot place CSS out of head section (%s)" % placement)


class _Embed(_DocumentVisitor):

    def act(self, page_doc, _):
        content = self.read_method()
        self.embed(page_doc, content)


class _ExportToTargetFs(_DocumentVisitor):
    resource_subdir = "resources"

    def act(self, page_doc, yawrap_instance):
        self._check_file_name_provided()
        href = self._create_local_file(yawrap_instance)
        self.link(page_doc, href)

    def _check_file_name_provided(self):
        if not self.file_name:
            raise ValueError("You need to provide filename in order to store "
                             "the content for %s operation." % self.__class__.__name__)

    def _create_local_file(self, yawrap_instance):
        root_dir = yawrap_instance.get_root_dir()
        target_file = os.path.join(root_dir, self.resource_subdir, self.file_name)
        content = self.read_method()
        self._save_as_file(content, target_file)
        href = posixpath.relpath(target_file, yawrap_instance._target_dir)
        return href

    @staticmethod
    def _save_as_file(str_content, target_file_path):
        if os.path.exists(target_file_path):
            warn_("File: %s already exists, overwritting." % target_file_path)
        with open(make_place(target_file_path), "wt") as ff:
            ff.write(str_content)


class _LinkExternalURL(_DocumentVisitor):

    def __init__(self, url, placement=HEAD):
        assert is_url(url), "That doesn't seem to be a valid url: %s" % url
        self.url = url
        super(_LinkExternalURL, self).__init__(lambda: self.url, placement=placement)

    def act(self, page_doc, _):
        self.link(page_doc, self.url)

    @classmethod
    def from_url(cls, url, placement=HEAD):
        return cls(url, placement)  # constructor bypass

    @classmethod
    def from_file(cls, *_, **__):
        raise TypeError("Cannot reference remote/external file by local file content.")


class EmbedCss(_Embed, _CssResource):
    pass


class EmbedJs(_Embed, _JsResource):
    pass


class LinkCss(_ExportToTargetFs, _CssResource):
    pass


class LinkJs(_ExportToTargetFs, _JsResource):
    pass


class ExternalCss(_LinkExternalURL, _CssResource):
    pass


class ExternalJs(_LinkExternalURL, _JsResource):
    pass
